import json
import tabulate
from operator import itemgetter
import os
import datetime
import re
import hashlib
import yaml
import datetime
import requests

# Currently, the supported photon version is 4
PHOTON_VERSION = os.environ.get("PHOTON_VERSION", 4)

time_of_scan = datetime.datetime.now()
scanner = None
scanner_options = {"cbctl": "cbctl", "twistcli": "twistcli"}

html_file_obj = None
RULES_FILE = "rules.yaml"
rules = None
exceptions = {
    "by_id": [],
    "by_package": [],
    "by_file": [],
}

# Populated during binary parsing. Used in the final json report
package_info = []

FIPS_STATUS = {
    0: "",
    1: "unknown",
    2: "\u001b[31;1mno\u001b[0m",
    3: "capable",
    4: "\u001b[32;1menforced\u001b[0m",
}

fips_checks = {
    "openssl": None,
    "java": None,
    "python": None,
    "golang": None,
    "nodejs": None,
}

malware_info = []

# Load the BASE_DIFF flag making sure we have a string (if not set)
base_diff_flag = str(os.environ.get("BASE_DIFF")).lower()
discovered_base = None


def is_exception_applicable(ex, image=None):
    """Return whether this exception is applicable to be loaded."""
    if image is None:
        if "IMAGE" in os.environ:
            image = os.environ["IMAGE"]

    # Check if exception is meant for this image
    if not ("image" in ex and re.search(ex["image"], image)):
        return False

    # Check the expiry of the exception
    now = datetime.datetime.now()
    until = now
    if "until" in ex:
        try:
            # Format is 2023-06-27
            until = datetime.datetime.strptime(ex["until"], "%Y-%m-%d")
        except Exception as e:
            # Could not parse. Default to now.
            pass

    if now > until:
        # Do not load this exception. It has expired
        return False

    return True


def load_rules():
    global rules

    if rules:
        # Rules already loaded
        return True

    try:
        with open(RULES_FILE, "r") as f:
            rules = yaml.safe_load(f)

        # Only add the exceptions for the current target image
        for ex in rules["exceptions"]:
            if is_exception_applicable(ex):
                # Exceptions by id e.g. CVE-0000-0000
                if "by_id" in ex:
                    exceptions["by_id"] = exceptions["by_id"] + ex["by_id"]

                # Exceptions by package e.g. {"package": ".*databind.*", version: ".*"}
                if "by_package" in ex:
                    exceptions["by_package"] = (
                        exceptions["by_package"] + ex["by_package"]
                    )

                # Exceptions by file/path e.g. "my-amazing-package-.*.jar"
                if "by_file" in ex:
                    exceptions["by_file"] = exceptions["by_file"] + ex["by_file"]

    except Exception as e:
        # We should use a logger here
        print("Something went wrong loading the exceptions.", e)
        return False

    return True


def is_exception(v):
    # Check by id
    if "id" in v:
        for ex in exceptions["by_id"]:
            if re.search(ex, v["id"]):
                return True

    # Check by package/library
    if "package_name" in v and "package_version" in v:
        for ex in exceptions["by_package"]:
            if re.search(ex["package"], v["package_name"]) and re.search(
                ex["version"], v["package_version"]
            ):
                return True

    # Check by file
    if "paths" in v:
        for ex in exceptions["by_file"]:
            for p in v["paths"]:
                if re.search(ex, p):
                    return True

    return False


def report_diff(current_report={}, baseline_report={}):
    new_report = []

    for entry in current_report:
        # Make a new copy, so that the original is not changed
        new_entry = entry.copy()

        if new_entry in baseline_report and base_diff_flag == "true":
            new_entry["status"] = vuln_status["base"]
        new_report.append(new_entry)

    return new_report


def calculate_report(full_report={}, baseline_report={}):
    vulns = {
        "target_all": None,
        "target_with_fixes": None,
        "baseline_all": None,
        "baseline_with_fixes": None,
    }
    vulns["target_all"] = {
        "critical": 0,
        "high": 0,
        "medium": 0,
        "low": 0,
        "unknown": 0,
        "total": 0,
    }
    vulns["target_with_fixes"] = {
        "critical": 0,
        "high": 0,
        "medium": 0,
        "low": 0,
        "unknown": 0,
        "total": 0,
    }
    vulns["baseline_all"] = {
        "critical": 0,
        "high": 0,
        "medium": 0,
        "low": 0,
        "unknown": 0,
        "total": 0,
    }
    vulns["baseline_with_fixes"] = {
        "critical": 0,
        "high": 0,
        "medium": 0,
        "low": 0,
        "unknown": 0,
        "total": 0,
    }

    reports = {"target": full_report, "baseline": baseline_report}
    for r in reports:
        for item in reports[r]:
            severity = "UNKNOWN"
            has_fix = 0

            # Exception check:
            #   item["is_exception"] comes from CBC
            #   item["status"] = vuln_status["exception"] comes `rules.yaml`
            if ("is_exception" in item and item["is_exception"]) or (
                "status" in item and item["status"] == vuln_status["exception"]
            ):
                # this is an exception skip this item
                continue

            if "status" in item and item["status"] == vuln_status["base"]:
                # This is an entry in the target image coming from the base image, skip this item
                continue

            if "cvss" in item and "v3" in item["cvss"]:
                score = item["cvss"]["v3"]
                if score >= CVSS_RATINGS["critical"]["min"]:
                    severity = "CRITICAL"
                elif score >= CVSS_RATINGS["high"]["min"]:
                    severity = "HIGH"
                elif score >= CVSS_RATINGS["medium"]["min"]:
                    severity = "MEDIUM"
                else:
                    severity = "LOW"
            else:
                if "severity" in item and item["severity"].lower() in [
                    "critical",
                    "high",
                    "medium",
                    "low",
                    "unknown",
                ]:
                    severity = item["severity"]

            # This has been normalised in normalize_report()
            if "fix_available" in item and item["fix_available"] != "":
                has_fix = 1

            vulns[f"{r}_all"][severity.lower()] += 1
            vulns[f"{r}_with_fixes"][severity.lower()] += has_fix

    for v in ("target_all", "target_with_fixes", "baseline_all", "baseline_with_fixes"):
        total = 0
        for severity in ("Critical", "High", "Medium", "Low", "Unknown"):
            total += vulns[v][severity.lower()]

        vulns[v]["total"] = total

    return vulns


def check_image_base(json_target_file=None, json_base_file=None):
    global discovered_base
    # Check if the target image is based on the specified base image
    json_target = None
    json_base = None
    target_layers = None
    base_layers = None

    if not json_target_file or not json_base_file:
        # json files not provided
        return False

    try:
        with open(json_target_file, "r") as f:
            json_target = json.load(f)
    except Exception as e:
        json_target = {}

    try:
        with open(json_base_file, "r") as f:
            json_base = json.load(f)
    except Exception as e:
        json_base = {}

    json_c = json_target
    if len(json_c) > 0 and "RootFS" in json_c[0] and "Layers" in json_c[0]["RootFS"]:
        target_layers = json_c[0]["RootFS"]["Layers"]

    json_c = json_base
    if len(json_c) > 0 and "RootFS" in json_c[0] and "Layers" in json_c[0]["RootFS"]:
        base_layers = json_c[0]["RootFS"]["Layers"]

    if not target_layers or not base_layers:
        # could not extract layers for both images
        return False

    # check if target is based on the base image
    if base_layers[-1] not in target_layers:
        # Target image is not based on the specified base or we couldn't
        if (
            "BASE" in os.environ
            and os.environ["BASE"]
            and "BASE_IMAGES" in os.environ
            and os.environ["BASE_IMAGES"]
        ):
            approved = os.environ["BASE"]

            if "BASE_IMAGES" in os.environ:
                approved = (
                    " " + os.environ["BASE_IMAGES"]
                )  # Added ' ' prefix for splitting below

                status_message["info"][
                    0
                ] = "Target image not based on any of the approved latest RE base images:{}".format(
                    approved.replace(" ", "\n  ")
                )
                scan_status["info"] = True

    else:
        if "BASE" in os.environ:
            discovered_base = os.environ["BASE"]

    return base_layers[-1] in target_layers


def show_malware_table(html_file=None):
    hdrs = ["Malware Finding", "Additional Info"]
    rows = []

    if len(malware_info) > 0:
        for _dict in malware_info:
            finding = [_dict["finding"], _dict["additional_info"]]
            rows.append(finding)
    else:
        rows.append(["No malware found", "No additional info"])

    print(tabulate.tabulate(rows, headers=hdrs, tablefmt="grid"))
    html_output(html_file, tabulate.tabulate(rows, headers=hdrs, tablefmt="unsafehtml"))


def get_version_from_os_info(s=""):
    """Return Major.Minor from os info or -1.0."""
    last_word = s.split(" ")[-1]
    try:
        version = (float)(".".join(last_word.split(".")[:2]))
    except BaseException as e:
        version = -1.0

    return version


def show_info_os_table(html_file=None, info_os_string=""):
    if info_os_string is None:
        info_os_string = ""

    hdrs = ["OS Release", "Status"]

    status = colour_if_non_zero("green", "Supported")

    if info_os_string == "":
        status = "Unable to get OS info"

    elif "vmware photon" not in info_os_string.lower():
        status = colour_if_non_zero("yellow", "Not Photon Based")

    elif get_version_from_os_info(info_os_string) < PHOTON_VERSION:
        status = colour_if_non_zero("red", "Deprecated")

    # Adding `+ 0.999` to accomodate for minor versions (even if not currently seen in photon)
    elif get_version_from_os_info(info_os_string) > PHOTON_VERSION + 0.999:
        status = colour_if_non_zero("yellow", "Not supported yet")

    rows = [[info_os_string, status]]

    print(tabulate.tabulate(rows, headers=hdrs, tablefmt="grid"))
    html_output(html_file, tabulate.tabulate(rows, headers=hdrs, tablefmt="unsafehtml"))


def show_fips_table(html_file=None):
    hdrs = ["FIPS Check", "FIPS Mode"]
    rows = []

    for c in fips_checks.keys():
        if fips_checks[c]:
            rows.append([c, FIPS_STATUS[fips_checks[c]]])

    print(tabulate.tabulate(rows, headers=hdrs, tablefmt="grid"))
    if rows == []:
        # There is an issue with tabulate which causes it to not display the
        # table properly if it has no 'body'
        rows = [["", ""]]
    html_output(html_file, tabulate.tabulate(rows, headers=hdrs, tablefmt="unsafehtml"))


def show_vuln_table(full_report={}, html_file=None):
    # Print our custom table

    hdrs = ["VULN ID", "STATUS", "TYPE", "Package", "Fix Available", "CVSS", "Paths"]
    rows = []

    for v in full_report:
        f_colour = lambda msg: msg

        if v["status"] == vuln_status["fail"]:
            f_colour = lambda msg: colour_if_non_zero("red", msg)

        if v["status"] == vuln_status["warning"] or v["status"] == vuln_status["base"]:
            f_colour = lambda msg: colour_if_non_zero("yellow", msg)

        entry = [
            v["id"],
            f_colour(v["status"]),
            v["package_type"],
            f'{v["package_name"]} {v["package_version"]}',
            v["fix_available"],
            v["cvss"]["v3"] if "cvss" in v and "v3" in v["cvss"] else -1,
            "\n".join(v["paths"]) if "paths" in v else "",
        ]

        if entry in rows:
            # skip duplicates (from other generic results)
            continue

        rows.append(entry)

    # Sort by CVSS
    rows = sorted(rows, key=itemgetter(5), reverse=True)

    # Actually print the table
    # print(tabulate.tabulate(rows, headers=hdrs, tablefmt="grid")) # Skip in CLI output
    html_output(html_file, tabulate.tabulate(rows, headers=hdrs, tablefmt="unsafehtml"))


def colour_if_non_zero(c, msg):
    colour_pre = ""
    colour_post = ""

    if msg == "0" or msg == 0:
        return msg

    if c.lower() == "red":
        colour_pre = "\u001b[31;1m"
        colour_post = "\u001b[0m"

    if c.lower() == "yellow":
        colour_pre = "\u001b[33;1m"
        colour_post = "\u001b[0m"

    if c.lower() == "green":
        colour_pre = "\u001b[32;1m"
        colour_post = "\u001b[0m"

    return colour_pre + str(msg) + colour_post


def init_html_file(html_file=None):
    global html_file_obj

    try:
        html_file_obj = open(html_file, "w")
    except Exception as e:
        return False

    # What follows is a really dirty way to put html

    bootstrap_css = '<link href="https://cdn.jsdelivr.net/npm/bootstrap@5.1.3/dist/css/bootstrap.min.css" rel="stylesheet" integrity="sha384-1BmE4kWBq78iYhFldvKuhfTAU6auU8tT94WrHftjDbrCEXSU1oBoqyl2QvZ6jIW3" crossorigin="anonymous">'
    bootstrap_js = '<script src="https://cdn.jsdelivr.net/npm/bootstrap@5.1.3/dist/js/bootstrap.bundle.min.js" integrity="sha384-ka7Sk0Gln4gmtz2MlQnikT1wXgYsOg+OMhuP+IlRH9sENBO0LRn5q+8nbTov4+1p" crossorigin="anonymous"></script>'

    print(bootstrap_css, file=html_file_obj)
    print(bootstrap_js, file=html_file_obj)

    item_id = "1"
    item_name = "Scanned Image"
    button_text = "bg-default"

    if "IMAGE" in os.environ:
        item_name = os.environ["IMAGE"]

        # Use unique random identifier for the html tags
        item_id = hashlib.sha256(os.urandom(16)).hexdigest()

    if scan_status["fail"]:
        button_text = "text-danger"
        status_hint = "fail"
    elif scan_status["warning"]:
        button_text = "text-warning"
        status_hint = "warn"
    elif scan_status["info"]:
        button_text = "text-info"
        status_hint = "info"
    else:
        button_text = "text-success"
        status_hint = "pass"

    print("<div class='accordion-item'>", file=html_file_obj)
    print(f"\t<h2 class='accordion-header' id='heading-{item_id}'>", file=html_file_obj)
    print(
        f"\t\t<button class='accordion-button bg-light {button_text}' type='button' data-bs-toggle='collapse' data-bs-target='#collapse-{item_id}' aria-expanded='true' aria-controls='collapse-{item_id}'> <span class='text-muted'> {scanner}: &emsp; </span> <span style=\"font-size:1rem; font-family: monospace\"> ({status_hint}) {item_name} </span> <span class='text-muted'> &emsp; (click to expand) </span> </button>",
        file=html_file_obj,
    )
    print("\t</h2>", file=html_file_obj)
    print(
        f"\t<div id='collapse-{item_id}' class='accordion-collapse collapse show' aria-labelledby='heading' data-bs-parent='#accordionExample'>",
        file=html_file_obj,
    )
    print("\t\t<div class='accordion-body'>", file=html_file_obj)
    return True


def close_html_file():
    if html_file_obj is None:
        return

    print("\t\t</div>", file=html_file_obj)
    print("\t</div>", file=html_file_obj)
    print("</div>", file=html_file_obj)

    html_file_obj.close()


def html_output(html_file=None, data=None):
    global html_file_obj

    if not html_file:
        # nothing to do
        return

    if html_file_obj is None and not init_html_file(html_file):
        # something went wrong
        return

    # We should be good now

    data = data.replace("<table>", "<table class='table table-hover w-auto'>")
    data = data.replace("<thead>", "<thead class='table-light'>")

    # Replace prefix for red
    data = data.replace("\u001b[31;1m", "<p class='text-danger'>")

    # Replace prefix for yellow
    data = data.replace("\u001b[33;1m", "<p class='text-warning'>")

    # Replace prefix for green
    data = data.replace("\u001b[32;1m", "<p class='text-success'>")

    # Replace suffix
    data = data.replace("\u001b[0m", "</p>")

    # These are from the summary table's notes
    data = data.replace("\n    ", "</br>&emsp;&emsp;")
    data = re.sub(
        r"([^<>])\n", r"\1</br>", data
    )  # Add html </br>, but not for '\n' after tags (from tabulate's output)

    # Make URIs clickable links
    uri_regex = r"(https?:\/\/(www\.)?[-a-zA-Z0-9@:%._\+~#=]{1,256}\.[a-zA-Z0-9()]{1,6}\b([-a-zA-Z0-9()!@:%_\+.~#?&\/\/=]*))"
    data = re.sub(uri_regex, r'<a href="\1"> \1 </a>', data)

    # Add links to NVD for CVEs
    data = re.sub(
        "(CVE-[0-9]+-[0-9]+)",
        r'<a href="https://nvd.nist.gov/vuln/detail/\1"> \1 </a>',
        data,
    )

    print(data, file=html_file_obj)
    html_file_obj.flush()
    return


def parse_license(s):
    # There are non-consistent license entries:
    #   "GPLv2+ and GPLv3+ and LGPLv2+"
    #   "LGPLv2.1,GPLv2"
    #   "GPLv2+ or LGPLv3+"
    #   "The Apache Software License, Version 2.0"

    s = re.sub(" and ", "_SEPARATOR_", s)
    s = re.sub(" or ", "_SEPARATOR_", s)
    s = re.sub(r",([^ ])", r"_SEPARATOR_\1", s)

    return s.split("_SEPARATOR_")


def normalize_output(normalized_report):
    scan_report = {}
    scan_report["metadata"] = {}
    scan_report["vulnerabilities"] = []
    scan_report["packages"] = []
    scan_report["metadata"]["image"] = ""
    scan_report["metadata"]["scan_time"] = time_of_scan.isoformat()
    scan_report["metadata"]["scanner"] = ""
    scan_report["metadata"]["status"] = scan_status

    if "IMAGE" in os.environ:
        scan_report["metadata"]["image"] = os.environ["IMAGE"]

    if scanner:
        scan_report["metadata"]["scanner"] = scanner

    # Add vulnerability information
    for v in normalized_report:
        # Entry with default values
        e = {}
        e["id"] = ""
        e["package_name"] = ""
        e["package_version"] = ""
        e["package_type"] = ""
        e["fix_available"] = ""
        e["cvss_score"] = None
        e["fix_date"] = ""
        e["paths"] = []
        e["status"] = ""

        # Only use the predefined values for the normalized json report
        if "id" in v:
            e["id"] = v["id"]

        if "package_name" in v:
            e["package_name"] = v["package_name"]

        if "package_version" in v:
            e["package_version"] = v["package_version"]

        if "package_type" in v:
            e["package_type"] = v["package_type"]

        if "fix_available" in v:
            e["fix_available"] = v["fix_available"]

        if "cvss" in v and "v3" in v["cvss"]:
            e["cvss_score"] = v["cvss"]["v3"]

        if "fix_date" in v:
            e["fix_date"] = v["fix_date"]

        if "paths" in v:
            e["paths"] = v["paths"]

        if "status" in v:
            e["status"] = v["status"]

        scan_report["vulnerabilities"].append(e)

    # Add package/library information
    for p in package_info:
        # Entry with default values
        e = {}
        e["name"] = ""
        e["version"] = ""
        e["type"] = ""
        e["licenses"] = []
        e["path"] = ""

        # Only use the predefined values for the normalized json report
        if "name" in p:
            e["name"] = p["name"]

        if "version" in p:
            e["version"] = p["version"]

        if "type" in p:
            e["type"] = p["type"]

        if "licenses" in p:
            if isinstance(p["licenses"], list):
                if p["licenses"] == []:
                    # in CBC this is oddly empty usually. Data is somewhere else
                    if "metadata" in p and "license" in p["metadata"]:
                        e["licenses"].append(parse_license(p["metadata"]["license"]))
                else:
                    # We have a list of licenses
                    for li in p["licenses"]:
                        e["licenses"] = e["licenses"] + parse_license(li)

            elif isinstance(p["licenses"], str):
                # We have a string license
                e["licenses"] = e["licenses"] + parse_license(p["licenses"])

        # This is for twistcli
        if "path" in p:
            e["path"] = p["path"]

        # This is for cbctl
        if (
            "locations" in p
            and isinstance(p["locations"], list)
            and len(p["locations"]) > 0
            and "path" in p["locations"][0]
        ):
            e["path"] = p["locations"][0]["path"]

        scan_report["packages"].append(e)

    return scan_report


def save_json_output(normalized_output={}, json_file=None):
    try:
        f = open(json_file, "w")
        json.dump(normalized_output, f)
        f.close()
    except Exception as e:
        # We should use a logger here
        print(f"Error saving JSON output to {json_file}")
        return False

    return True


def show_report(full_report={}, baseline_report={}, html_file=None):
    vulns = calculate_report(full_report, baseline_report)
    hdrs = [
        "Vulnerability Severity",
        "BASELINE\nAll",
        "BASELINE\nWith Fixes",
        "DEVIATION FROM BASELINE\nAll",
        "DEVIATION FROM BASELINE\nWith Fixes",
    ]
    rows = []

    for severity in ("Critical", "High", "Medium", "Low", "Unknown", "Total"):
        colour = lambda c, msg: msg
        if severity in ["Critical", "High"]:
            colour = colour_if_non_zero

        rows.append(
            [
                severity,
                vulns["baseline_all"][severity.lower()],
                vulns["baseline_with_fixes"][severity.lower()],
                colour("yellow", vulns["target_all"][severity.lower()]),
                colour("red", vulns["target_with_fixes"][severity.lower()]),
            ]
        )

    # Remove baseline and fix "deviation" message if
    # baseline_diff_flag != "true"

    if base_diff_flag != "true":
        hdrs = ["Vulnerability Severity", "All", "With Fixes"]
        rows = list(map(lambda row: [row[0], row[3], row[4]], rows))

    print(tabulate.tabulate(rows, headers=hdrs, tablefmt="grid"))
    html_output(html_file, tabulate.tabulate(rows, headers=hdrs, tablefmt="unsafehtml"))

    hdrs = ["Status", "Notes"]
    rows = []

    rows.append(["Scan Time", time_of_scan.isoformat()])

    if scanner:
        rows.append(["Scanner", scanner])

    # This will be the deduced base image or None
    if base_diff_flag == "true":
        rows.append(["Discovered Base", discovered_base])

    if "BASE" in os.environ and base_diff_flag == "true":
        rows.append(["Selected Baseline", os.environ["BASE"]])

    if "IMAGE" in os.environ:
        rows.append(["Target Image", os.environ["IMAGE"]])

    if "SAAS_BASE_URL" in os.environ and "image_sha256" in os.environ:
        rows.append(
            [
                "CBC Image Entry",
                f"https://{os.environ['SAAS_BASE_URL']}/kubernetes/image/{os.environ['image_sha256']}/overview",
            ]
        )

    if "JENKINS_BUILD_URL" in os.environ:
        rows.append(["Final Scan Report", f"{os.environ.get('JENKINS_BUILD_URL')}artifact/{os.environ.get('REP_PATH_CLI')}/final_report.html"])

    if scan_status["fail"]:
        rows.append(
            [
                "\u001b[31;1mTEST FAILED\u001b[0m",
                "".join(status_message["fail"].values())
                + "\n\nInstructions available at https://confluence.eng.vmware.com/x/bgClQ",
            ]
        )
    else:
        rows.append(["\u001b[32;1mTEST PASSED\u001b[0m", "", ""])

    if scan_status["warning"]:
        rows.append(
            [
                "\u001b[33;1mWARNING\u001b[0m",
                "".join(status_message["warning"].values())
                + "\n\nDeviation Request is needed for FedRAMP",
            ]
        )

    if scan_status["info"]:
        rows.append(["INFO", "".join(status_message["info"].values()), ""])

    print(tabulate.tabulate(rows, headers=hdrs, tablefmt="grid"))
    html_output(html_file, tabulate.tabulate(rows, headers=hdrs, tablefmt="unsafehtml"))

    close_html_file()

    return not scan_status["fail"]


vuln_status = {
    "fail": "Fail",
    "pass": "Pass",
    "warning": "Warn",
    "exception": "Except",
    "base": "Base",
}

scan_status = {"fail": False, "warning": False, "info": False}

# Check report_message for more info on priorities
status_message = {
    "fail": {  # This is to sort the messages e.g. Critical vulns is always 1st
        -1: "",  # Malware found
        0: "",
        1: "",
        2: "",
        3: "",
        4: "",  # FIPS if failing
    },
    "warning": {  # This is to sort the messages e.g. Critical vulns is always 1st
        0: "",
        1: "",
        2: "",
    },
    "info": {  # This is to sort the messages e.g. Critical vulns is always 1st
        0: "",
        1: "",
        2: "",
        4: "",  # FIPS if non-failing
    },
}

CVSS_RATINGS = {
    "none": {"min": 0.0, "max": 0.0},
    "low": {"min": 0.1, "max": 3.9},
    "medium": {"min": 4.0, "max": 6.9},
    "high": {"min": 7.0, "max": 8.9},
    "critical": {"min": 9.0, "max": 10.0},
}

# SLOs are 30 days shorter than FedRAMP's to allow time for fixing
SLO = {
    "none": float("inf"),  # i.e. always allowed
    "low": 150,
    "medium": 60,
    "high": float("-inf"),  # i.e. not allowed
    "critical": float("-inf"),  # i.e. not allowed
}

report_message = {
    "fail": {
        "critical": {
            "priority": 0,
            "msg": f"Critical severity vulnerabilities\n    (CVSSv3 score in [{CVSS_RATINGS['critical']['min']} ; {CVSS_RATINGS['critical']['max']}])\n",
        },
        "high": {
            "priority": 1,
            "msg": f"High severity vulnerabilities\n    (CVSSv3 score in [{CVSS_RATINGS['high']['min']} ; {CVSS_RATINGS['high']['max']}])\n",
        },
        "medium": {
            "priority": 2,
            "msg": f"Medium severity vulnerabilities missed SLO\n    (CVSSv3 score in [{CVSS_RATINGS['medium']['min']} ; {CVSS_RATINGS['medium']['max']}])\n    SLO: {SLO['medium']}d\n",
        },
        "low": {
            "priority": 3,
            "msg": f"Low severity vulnerabilities missed SLO\n    (CVSSv3 score in [{CVSS_RATINGS['low']['min']} ; {CVSS_RATINGS['low']['max']}])\n    SLO: {SLO['low']}d\n",
        },
    },
    "warning": {
        "critical": {
            "priority": 0,
            "msg": f"Critical severity vulnerabilities without fixes\n    (CVSSv3 score in [{CVSS_RATINGS['critical']['min']} ; {CVSS_RATINGS['critical']['max']}])\n",
        },
        "high": {
            "priority": 1,
            "msg": f"High severity vulnerabilities without fixes\n    (CVSSv3 score in [{CVSS_RATINGS['high']['min']} ; {CVSS_RATINGS['high']['max']}])\n",
        },
        "medium": {
            "priority": 2,
            "msg": f"Medium severity vulnerabilities without fixes missed SLO\n    (CVSSv3 score in [{CVSS_RATINGS['medium']['min']} ; {CVSS_RATINGS['medium']['max']}])\n    SLO: {SLO['medium']}d\n",
        },
        "low": {
            "priority": 3,
            "msg": f"Low severity vulnerabilities without fixes missed SLO\n    (CVSSv3 score in [{CVSS_RATINGS['low']['min']} ; {CVSS_RATINGS['low']['max']}])\n    SLO: {SLO['low']}d\n",
        },
    },
}


def analyze_vulns(full_report):
    # Generic (non-mapped from binary to vulns) results
    for v in full_report:
        from_base = False

        # Check for exceptions
        if is_exception(v):
            v["status"] = vuln_status["exception"]

            # There is an exception for this entry, shall not be processed any further
            continue

        if "status" in v and v["status"] == vuln_status["base"]:
            # This comes from the base image
            from_base = True

        # Decistion will be made further down
        v["status"] = vuln_status["pass"]

        # Deduce CVSS rating (e.g. "critical")
        cvss_rating = "none"
        if "cvss" in v and "v3" in v["cvss"]:
            cvss = v["cvss"]["v3"]

            # cvss above is a number, converting that to a rating (i.e. "critica", "high" etc)
            for k in CVSS_RATINGS:
                if CVSS_RATINGS[k]["min"] <= cvss and cvss <= CVSS_RATINGS[k]["max"]:
                    cvss_rating = k
                    break

        # Check if fix is available
        fix_available = False
        if "fix_available" in v and v["fix_available"] != "":
            fix_available = True

        # Calculate elapsed time since fix is available
        try:
            # Try to get and parse v["fix_date"]
            date_fix = datetime.datetime.strptime(v["fix_date"], "%Y-%m-%d")
        except Exception as e:
            # Could not parse date, just use today
            date_fix = datetime.datetime.now()

        elapsed_days = (datetime.datetime.now() - date_fix).days

        decision = None
        if elapsed_days > SLO[cvss_rating] and fix_available:
            if from_base:
                # Only warn since the vulnerability is in the base image
                decision = "warning"
            else:
                decision = "fail"

        if elapsed_days > SLO[cvss_rating] and not fix_available:
            decision = "warning"

        if decision:
            prio = report_message[decision][cvss_rating]["priority"]
            msg = report_message[decision][cvss_rating]["msg"]

            # mark the vulnerability entry as fail/warning
            if from_base:
                v["status"] = vuln_status["base"]
            else:
                v["status"] = vuln_status[decision]

            # Add message for the summary table e.g. "Critical severity vulnerabilities"
            status_message[decision][prio] = msg

            # Mark overall status flag as fail/warning
            scan_status[decision] = True


def parse_twistlock(twistlock_json={}):
    """Parse Twistlock results."""
    global scanner
    scanner = scanner_options["twistcli"]

    report = []

    if (
        "results" in twistlock_json
        and type(twistlock_json["results"]) is list
        and len(twistlock_json["results"]) > 0
    ):
        # Use the first result. There should not be more in our use case.
        result_json = twistlock_json["results"][0]
    else:
        return report

    if "vulnerabilities" not in result_json:
        return report

    packages = []
    if "packages" in result_json:
        packages = result_json["packages"]

    for v in result_json["vulnerabilities"]:
        parsed = {}
        if "id" in v:
            parsed["id"] = v["id"]
        if "packageName" in v:
            pname = v["packageName"]
            parsed["package"] = pname
            parsed["package_name"] = pname
            if "packageVersion" in v:
                parsed["package"] += "-" + v["packageVersion"]
                parsed["package_version"] = v["packageVersion"]

            # Get the (first) item from packges by its name and version
            pkg = None
            for check_pkg in packages:
                if (
                    check_pkg["name"] == parsed["package_name"]
                    and check_pkg["version"] == parsed["package_version"]
                ):
                    pkg = check_pkg
                    break

            parsed["package_type"] = ""
            if pkg and "type" in pkg:
                parsed["package_type"] = pkg["type"]
        if "status" in v:
            parsed["fix_available"] = v["status"]
        else:
            parsed["fix_available"] = ""

        parsed["cvss"] = {"v3": -1}
        if "cvss" in v and "vector" in v:
            if v["vector"][:6] == "CVSS:3":
                parsed["cvss"] = {"v3": v["cvss"]}
            if v["vector"][:6] == "CVSS:2":
                parsed["cvss"] = {"v2": v["cvss"]}
        elif "cvss" in v:
            # We have a CVSS score, but not the CVSS vector
            # We shall assume CVSSv3
            parsed["cvss"] = {"v3": v["cvss"]}

        if "severity" in v:
            parsed["severity"] = v["severity"]

        if "fixDate" in v:
            parsed["fix_date"] = v["fixDate"].split("T")[0]
        else:
            parsed["fix_date"] = ""

        report.append(parsed)

    # Enrich with binary data
    report = parse_twistcli_binary_vulnerabilities(report, twistlock_json)

    return report


def parse_twistcli_binary_vulnerabilities(report, twistlock_json):
    scan_results = []
    global package_info

    # Unpack results
    if (
        "results" in twistlock_json
        and isinstance(twistlock_json["results"], list)
        and len(twistlock_json["results"]) == 1
    ):
        scan_results = twistlock_json["results"][0]

    if "vulnerabilities" not in scan_results or (
        "applications" not in scan_results and "packages" not in scan_results
    ):
        # Not enough data to process
        return report

    # This dictionary will later look like {('com.fasterxml.jackson.core_jackson-databind', '2.4.0'): ['/htrace-core4-4.1.0-incubating.jar']}
    paths_by_package = {}

    binaries = []

    # Add "applications" if found
    try:
        binaries = binaries + scan_results["applications"]
    except Exception as e:
        pass

    # Add "packages" if found
    try:
        binaries = binaries + scan_results["packages"]
    except Exception as e:
        pass

    package_info = binaries

    for b in binaries:
        if "name" not in b or "version" not in b or "path" not in b:
            # Nothing to do here
            continue

        t = (b["name"], b["version"])
        if t not in paths_by_package:
            paths_by_package[t] = []

        paths_by_package[t].append(b["path"])

    # Add vulnerability entries with "path"
    for v in list(report):
        if "package_name" in v and "package_version" in v and "path" not in v:
            t = (v["package_name"], v["package_version"])
            if t in paths_by_package:
                report.append(v | {"paths": paths_by_package[t]})

                # Remove the entry without path
                report.remove(v)

    return report


def parse_cbctl_binary_vulnerabilities(report, packages):
    global scanner
    global package_info
    scanner = scanner_options["cbctl"]

    if "artifacts" not in packages or packages["artifacts"] is None:
        # Nothing to parse
        return report

    # This dictionary will later look like {('jackson-databind', '2.4.0'): ['/htrace-core4-4.1.0-incubating.jar']}
    paths_by_package = {}

    package_info = packages["artifacts"]

    for art in packages["artifacts"]:
        if "name" not in art or "version" not in art or "locations" not in art:
            # Nothing to do here
            continue

        t = (art["name"], art["version"])

        if (art["name"], art["version"]) not in paths_by_package:
            paths_by_package[t] = []

        if "foundBy" in art and (
            "db-cataloger" in art["foundBy"] or "package-cataloger" in art["foundBy"]
        ):
            # This must have come from os packages e.g. apkdb, dpkgdb, rpmdb
            # Or from some other package manager e.g. javascript-package-cataloger, python-package-cataloger

            # For those we don't want to have all the files as separate entries
            # Just add one of those to help locate the path (e.g. for python, there
            # might be multiple virtual environments)
            paths_by_package[t].append(art["locations"][0]["path"])
            continue

        for loc in art["locations"]:
            if "path" in loc:
                paths_by_package[t].append(loc["path"])

    # Add vulnerability entries with "path"
    for v in list(report):
        if "package_name" in v and "package_version" in v and "path" not in v:
            t = (v["package_name"], v["package_version"])
            if t in paths_by_package:
                report.append(v | {"paths": paths_by_package[t]})

                # Remove the entry without path
                report.remove(v)

    return report


def normalize_report(report=[]):
    load_rules()

    for v in report:
        # Normalize the different messages referring to no fix available
        if "fix_available" in v:
            if v["fix_available"] in ["", "open"]:
                v["fix_available"] = ""
        else:
            v["fix_available"] = ""

        if "severity" in v:
            cvss_v3 = ""
            if "cvss" in v and "v3" in v["cvss"]:
                cvss_v3 = v["cvss"]["v3"]

            if cvss_v3 == "" or (float)(cvss_v3) <= 0.0:
                # We don't really have a CVSSv3 score
                # We will assign score based on the severity
                if v["severity"].lower() == "critical":
                    v["cvss"] = {"v3": 9.01}

                if v["severity"].lower() == "high":
                    v["cvss"] = {"v3": 7.01}

                if v["severity"].lower() == "medium":
                    v["cvss"] = {"v3": 4.01}

                if v["severity"].lower() == "low":
                    v["cvss"] = {"v3": 0.11}

        if "fix_date" in v:
            # Only use the date portion i.e. 2021-01-31T01:15:00Z -> 2021-01-31
            v["fix_date"] = v["fix_date"].split("T")[0]

    deduplicated_report = []
    for i in range(len(report)):
        # Only use the entries that are not duplicated later
        if report[i] not in report[i + 1:]:
            deduplicated_report.append(report[i])

    return deduplicated_report


def check_compliance_twistcli(compliances, notify: bool):
    # Twistlock uses IDs to categorize vulnerabilities, malware, miners, etc
    # ID 422: malware
    # ID 426: miners

    if len(compliances) > 0:
        SLACK_WEBHOOK = str(os.environ.get("PSEC_MALWARE_SCANNER_SLACK_WEBHOOK_CIS"))
        JOB_URL = os.environ.get("JENKINS_BUILD_URL")
        c_item_list = []
        _fail = False

        for _dict in compliances:
            if "id" in _dict and _dict["id"] in [422, 426]:
                c_item_dict = {}
                c_item_dict["id"] = _dict["id"]
                c_item_dict["title"] = _dict["title"]
                c_item_dict["cause"] = _dict["cause"]

                c_item_list.append(c_item_dict)
                malware_info.append(
                    {"finding": _dict["title"], "additional_info": _dict["cause"]}
                )
                _fail = True

        if _fail:
            _msg = "\n ".join(
                f"ID: {d['id']}, Title: {d['title']}, Cause: {d['cause']}\n"
                for d in c_item_list
            )
            scan_status["fail"] = True
            status_message["fail"][-1] = f"Malware/Crypto Miner found!\n"
            slack_msg = f"""
                            {{
                                'scanner': 'twistcli',
                                'job_url': '{JOB_URL}',
                                'malware_info': '{_msg}'
                            }}
                        """

            if notify:
                requests.post(SLACK_WEBHOOK, slack_msg)


def check_file_reputation_cbctl(fr, notify: bool):
    # Check file reputation a.k.a. "malware scanning"

    if len(fr) > 0:
        SLACK_WEBHOOK = str(os.environ.get("PSEC_MALWARE_SCANNER_SLACK_WEBHOOK_CIS"))
        JOB_URL = os.environ.get("JENKINS_BUILD_URL")
        fr_item_list = []
        _fail = False

        for _dict in fr:
            fr_item_dict = {}
            fr_item_dict["file_name"] = _dict["file_name"]
            fr_item_dict["path"] = _dict["path"]

            fr_item_list.append(fr_item_dict)
            malware_info.append(
                {"finding": _dict["file_name"], "additional_info": _dict["path"]}
            )
            _fail = True

        if _fail:
            _msg = "\n ".join(
                f"File name: {d['file_name']}, Path: {d['path']}\n"
                for d in fr_item_list
            )
            scan_status["fail"] = True
            status_message["fail"][-1] = f"Malware found!\n"
            slack_msg = f"""
                            {{
                                'scanner': 'cbctl',
                                'job_url': '{JOB_URL}',
                                'malware_info': '{_msg}'
                            }}
                        """

            if notify:
                requests.post(SLACK_WEBHOOK, slack_msg)
