#! /usr/bin/env python3

"""PSEC BlackDuck Scanner.

This script uses BlackDuck to perform licensing,
vulnerabilities, and binary analysis against files and docker images.
"""

import argparse
import os
import re
import sys
import base64
import subprocess
import shutil
import requests
import yaml
import boto3
from requests.auth import HTTPBasicAuth


def validate_variables():
    """Validates the presence of required environment variables.

    Raises:
        ValueError: if any required environment variable is not set.
    """

    required_vars = ["BCOM_BDUCK_URL", "BCOM_BDUCK_API_TOKEN"]

    for var in required_vars:
        if not os.environ.get(var):
            raise ValueError(f"** ERROR: {var} environment variable is not set.")


def setup_aws_client(service, credentials, region=None):
    """Helper function to set up an AWS client for a given service."""

    return boto3.client(
        service,
        region_name=region,
        aws_access_key_id=credentials["AccessKeyId"],
        aws_secret_access_key=credentials["SecretAccessKey"],
        aws_session_token=credentials["SessionToken"],
    )


def assume_role(sts_client, role_arn, session_name):
    """Helper function to assume an AWS role."""

    return sts_client.assume_role(RoleArn=role_arn, RoleSessionName=session_name)[
        "Credentials"
    ]


def login_to_ecr(image_name):
    """Main function to log into AWS ECR.

    Args:
        image_name (str): Image name used to extract login information from.
    """

    try:
        aws_account = image_name.split(".")[0]
        region = image_name.split(".")[3]

        # Set up the initial AWS clients with psec credentials
        os.environ["AWS_ACCESS_KEY_ID"] = os.environ["AWS_SECPROD_KEY_ID"]
        os.environ["AWS_SECRET_ACCESS_KEY"] = os.environ["AWS_SECPROD_SECRET"]

        sts_client = boto3.client("sts")

        # Assume the psec role in SECPROD
        psec_role_arn = f"arn:aws:iam::{os.environ['AWS_SECPROD_ACCOUNT_ID']}:role/psec-ecr-image-scanner"
        secprod_credentials = assume_role(sts_client, psec_role_arn, "ecr-session")

        # Set up the client for the target account using the assumed role credentials
        target_account_role_arn = (
            f"arn:aws:iam::{aws_account}:role/psec-ecr-image-scanner"
        )
        target_sts_client = setup_aws_client("sts", secprod_credentials)

        target_credentials = assume_role(
            target_sts_client, target_account_role_arn, "ecr-session"
        )

        # Get the ECR credentials using boto3
        ecr_client = setup_aws_client("ecr", target_credentials, region)
        docker_password = ecr_client.get_authorization_token()["authorizationData"][0][
            "authorizationToken"
        ]
        docker_password = base64.b64decode(docker_password).decode().split(":")[1]

        # Log into ECR in the target AWS account
        login_command = [
            "docker",
            "login",
            "--username",
            "AWS",
            "--password-stdin",
            f"{aws_account}.dkr.ecr.{region}.amazonaws.com",
        ]

        subprocess.run(
            login_command,
            input=docker_password,
            text=True,
        )

        print("** INFO: Successfully logged into AWS ECR.")
    except Exception as e:
        print(f"** ERROR: An error occurred while logging into AWS ECR: {e}")
        sys.exit(0)


def download_file(url: str):
    """Downloads target file to be scanned.

    Args:
        url (str): URL of the file to be downloaded.

    Raises:
        RequestException: if the request fails.

    Returns:
        file_path (str): path to the file to be scanned.
    """

    session = requests.Session()

    try:
        auth = None

        if "usw1.packages.broadcom.com" in url:
            auth = HTTPBasicAuth(os.getenv("ARTIFACTORY_USER"), os.getenv("ARTIFACTORY_PASSWORD"))

        response = session.head(url, auth=auth)

        response.raise_for_status()

        file_name = url.split("/")[-1]
        file_path = os.path.join(os.getcwd(), file_name)

        with session.get(url, stream=True) as s_response:
            with open(file_path, "wb") as file:
                for chunk in s_response.iter_content(chunk_size=8192):
                    file.write(chunk)

        print(f"** INFO: File downloaded and saved at: {file_path}")
        return file_path
    except requests.exceptions.RequestException as e:
        print(f"** ERROR: {e}")
        return None


def get_projects(limit=1, parameters={}):
    headers = {}

    if limit:
        parameters.update({"limit": limit})

    url = f"{os.environ.get('BCOM_BDUCK_URL')}/{parameters}"
    headers["Accept"] = "application/vnd.blackducksoftware.project-detail-4+json"

    print(f"** INFO: Retrieving projects using url {url}")

    response = requests.get(url, headers=headers, verify=True)
    projects = response.json()

    return projects


def get_project_by_name(project_name):
    project_list = get_projects(parameters={f"q": "name:{project_name}"})

    for project in project_list["items"]:
        if project["name"] == project_name:
            return project


def read_yaml(project_groups, project_name: str):
    """Traverses the dictionary and returns the project name or None.

    Args:
        project_groups (dict): dict containing all project groups.
        project_name (str): the name of the project to search for.

    Returns:
        project_group (str): the project group name.
    """

    if isinstance(project_groups, dict):
        for key, value in project_groups.items():
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, dict):
                        result = read_yaml(item, project_name)

                        if result:
                            return result
                    elif project_name.lower() in item.lower():
                        return key
            elif isinstance(value, dict):
                result = read_yaml(value, project_name)

                if result:
                    return result
    return None


def get_project_group(project_name: str):
    """Determines the project group name.

    Searches for project_name in the mapping file.
    If no matches are found, project_name is set to a default value.

    Args:
        project_name (str): the name of the project to search for.

    Raises:
        Exception: if something fails.

    Returns:
        project_group (str): the project group name.
    """

    mapping_file = os.path.join(os.getcwd(), "blackduck/projects_mapping.yml")

    try:
        with open(mapping_file, 'r') as f:
            project_groups = yaml.safe_load(f)

        project_group = read_yaml(project_groups, project_name)

        if project_group is None:
            project_group = "CB Cloud (CBC)"
    except Exception as exc:
        print(f"** ERROR: An unexpected error occurred: {exc}")
        sys.exit(0)

    return project_group


def run_synopsys_detect(
    project_name: str,
    project_version: str,
    project_group: str,
    DETECT_TOOLS: str,
    EXCLUDED_TOOLS: str,
    target: str,
):
    """Runs BlackDuck Detect via shell command.

    Invokes BlackDuck Detect and scans a given target
    using several different detectors.

    Args:
        project_name (str): name of the project in BlackDuck HUB.
        project_version (str): version of the project in BlackDuck HUB.
        project_group (str): name of the project group in BlackDuck HUB.
        DETECT_TOOLS (str): Detect tools to include in the scan.
        EXCLUDED_TOOLS (str): Detect tools to exclude from the scan.
        target (str): file, directory or container image to be scanned.

    Raises:
        CalledProcessError: If the subprocess command fails.

    Returns:
        file_path (str): Path to the file to be scanned.
    """

    REP_PATH_CLI = os.environ.get("REP_PATH_CLI", ".")
    BCOM_BDUCK_URL = os.environ.get("BCOM_BDUCK_URL")
    BCOM_BDUCK_API_TOKEN = os.environ.get("BCOM_BDUCK_API_TOKEN")
    BDUCK_LOGGING_LEVEL = os.environ.get("BDUCK_LOGGING_LEVEL", "INFO")
    project_name = project_name
    project_version = project_version
    project_group = project_group or get_project_group(project_name)
    DETECT_TOOLS = DETECT_TOOLS or "BINARY_SCAN"
    EXCLUDED_TOOLS = EXCLUDED_TOOLS or "ALL"
    target = target
    image_name_pattern = re.compile(
        r"^([a-z0-9.-]+(:[a-zA-Z0-9.-]+)?(/[a-zA-Z0-9.-]+)+(:[a-zA-Z0-9.-]+)?)$"
    )

    cmd = [
        "java",
        "-jar",
        "/root/synopsys-detect.jar",
        f"--blackduck.url={BCOM_BDUCK_URL}",
        f"--blackduck.api.token={BCOM_BDUCK_API_TOKEN}",
        "--blackduck.trust.cert=true",
        f"--detect.project.group.name={project_group}",
        f"--detect.project.name={project_name.lower()}",
        f"--detect.project.version.name={project_version.lower()}",
        f"--detect.tools={DETECT_TOOLS}",
        f"--detect.tools.excluded={EXCLUDED_TOOLS}",
        f"--logging.level.com.synopsys.integration={BDUCK_LOGGING_LEVEL}",
        "--detect.risk.report.pdf=false",
        "--detect.wait.for.results=false",
        "--detect.policy.check.fail.on.severities=BLOCKER,CRITICAL",
        "--detect.cleanup=true",
        "--detect.docker.passthrough.shared.dir.path.local=/tmp/blackduck/",
    ]

    if os.path.isdir(target):
        cmd.append(f"--detect.source.path={target}")
    elif os.path.isfile(target):
        try:
            target = shutil.copy2(target, f"{REP_PATH_CLI}/")
            cmd.append(f"--detect.source.path={target}")
        except e:
            print(f"** ERROR: Failed to copy the target file: {e}")
            sys.exit(0)
    elif target.startswith(("http://", "https://")):
        target = download_file(target)

        try:
            target = shutil.copy2(target, f"{REP_PATH_CLI}/")
            cmd.append(f"--detect.source.path={target}")
        except e:
            print(f"** ERROR: Failed to copy the target file: {e}")
            sys.exit(0)
    elif image_name_pattern.match(target):
        pn_index = cmd.index(f"--detect.project.name={project_name}")
        pv_index = cmd.index(f"--detect.project.version.name={project_version}")
        pn = target.rsplit(":", 1)[0]
        pv = target.split(":")[-1]
        cmd[pn_index] = f"--detect.project.name={pn}"
        cmd[pv_index] = f"--detect.project.version.name={pv}"

        cmd.append(f"--detect.docker.image={target}")
        cmd.append(
            f"--detect.docker.passthrough.imageinspector.service.distro.default=centos"
        )

        if ".dkr.ecr." in target.lower():
            login_to_ecr(target)
    else:
        print("** ERROR: Provide a file URL, local file path, or container image name.")
        sys.exit(1)

    try:
        subprocess.run(cmd)
    except subprocess.CalledProcessError as e:
        print(f"** ERROR: Command execution failed: {e}")
        sys.exit(0)


def main():
    validate_variables()

    parser = argparse.ArgumentParser(description="PSEC BlackDuck Scanner")

    parser.add_argument(
        "-pn",
        "--project_name",
        help="Project name in BlackDuck",
        required=True,
        type=str,
    )
    parser.add_argument(
        "-pv",
        "--project_version",
        help="Project version in BlackDuck",
        required=True,
        type=str,
    )
    parser.add_argument(
        "-pg",
        "--project_group",
        help="Project parent group name in BlackDuck",
        required=False,
        type=str,
    )
    parser.add_argument(
        "-dt",
        "--detect_tools",
        help="Detect tools to include in the scan",
        required=False,
        type=str,
    )
    parser.add_argument(
        "-et",
        "--excluded_tools",
        help="Detect tools to exclude from the scan",
        required=False,
        type=str,
    )
    parser.add_argument(
        "-t",
        "--target",
        help="File URL, local file path, or full container image URL",
        required=True,
        type=str,
    )

    args = parser.parse_args()
    project_name = args.project_name
    project_version = args.project_version
    project_group = args.project_group
    DETECT_TOOLS = args.detect_tools
    EXCLUDED_TOOLS = args.excluded_tools
    target = args.target

    print("Inside the scanner.py file")
    print(f"name: {project_name}")
    print(f"version: {project_version}")
    print(f"group: {project_group}")
    print(f"target: {target}")

    print(f"** INFO: Initiating BlackDuck scan activities for {target} ..")

    for t in target.split():
        run_synopsys_detect(
            project_name,
            project_version,
            project_group,
            DETECT_TOOLS,
            EXCLUDED_TOOLS,
            t,
        )

    print("** INFO: BlackDuck scan activities are now complete.")


if __name__ == "__main__":
    main()
